#include<iostream>
#include<fstream>
#include<vector>
#include "ring.h"
#include "estimate.h"
#include "wheel.h"

#include "mpi.h"

using namespace std;


vector<Ring> r;
int cutoff;
int k;
ofstream outfile;
ofstream outfilemain;

void openoutfile(int k, int id)
{
  if(outfile.is_open()) outfile.close();
  char filename[]="parout-KKK-PPP.txt";
  filename[7]='0'+k/100;
  filename[8]='0'+(k/10)%10;
  filename[9]='0'+k%10;
  filename[11]='0'+id/100;
  filename[12]='0'+(id/10)%10;
  filename[13]='0'+id%10;
  outfile.open(filename);
}

void openoutfilemain(int k)
{
  if(outfilemain.is_open()) outfilemain.close();
  char filename[]="parout-KKK.txt";
  filename[7]='0'+k/100;
  filename[8]='0'+(k/10)%10;
  filename[9]='0'+k%10;
  outfilemain.open(filename);
}

void loadrings()
{
  char filename[]="xxx.txt";
  filename[0]='0'+k/100;
  filename[1]='0'+(k/10)%10;
  filename[2]='0'+k%10;

  ifstream f(filename);
  int numrings, rest;
  f >> numrings >> cutoff >> rest;
  r.resize(numrings);
  for(int i=0; i<numrings; i++)
  {
    int p,a,A;
    f >> p >> a >> A;
    r[i].init(p,k,a);
  }

}

bool filter(int128 x)
{
  for(int i=cutoff; i<r.size(); i++)
    if(!r[i].ones[x%r[i].m]) return false;
  return true;
}

int main(int argc, char* argv[])
{
  MPI_Init(&argc,&argv);
  int id, np;
  MPI_Comm_size(MPI_COMM_WORLD,&np);
  MPI_Comm_rank(MPI_COMM_WORLD,&id);
  //cout << "Enter k: ";
  //cin >> k;
  
  //cout << "k=" << k << endl;

  //for(k=10; k<=250; k++) {
  //k=275; // for testing
  for(k=378; k<=400; k++) 
  { // loop on different k values


  long double limit=estimateg(k);
  //outfile << "k="<<k << " Estimate="<<limit<<endl;

  //load the rings
  loadrings();

  // make the wheel
  Wheel w1, w2;
  int wheelsplit=w1.build1(r,cutoff,limit,25*np);
  w2.build2(r,wheelsplit,cutoff,limit,w1.m);

  if(id==0)
  {
    openoutfile(k,id);
    w1.print(outfile);
    w2.print(outfile);
  }

  int128 x;
  int128 x2;
  x=w1.first(k+2);
  int128 min=w2.max;
  int w1count=0;
  do {
    if(w1count%np==id) // we work on this x value
    { 
      // jump through w2 wheel values from this x
      x2=w2.first(x);
      do {
        if(x2>k+1 && x2<min && filter(x2)) { if(x2<min) w2.max=x2; min=x2; }
        if(w2.hasNext())
          x2=w2.next();
      } while(x2>0 && w2.hasNext());

    }
    x=w1.next();
    w1count++;
    // share minima
    if(w1count%np==0)
    {
    long double mymin=min*1.001; // round up!
    long double globalmin;
    MPI_Allreduce(&mymin,&globalmin,1,MPI_LONG_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
    if(globalmin<min)
      { w2.max=globalmin; min=globalmin; }
    }
  } while(x>0 && w1.hasNext());

/*
  x=w.first(k+2);
  do {
    if(x>k+1 && x<min && filter(x)) { if(x<min) w.max=x; min=x; }
    if(w.hasNext())
      x=w.next();
  } while(x>0 && w.hasNext());

  if(x>k+1 && filter(x)) { if(x<min) w.max=min=x; }
*/
  long double ldmin=min;
  long double reducemin;
  MPI_Allreduce(&ldmin,&reducemin,1,MPI_LONG_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
  //outfile << k << " " << min << endl;
  if(0.99999 < reducemin/ldmin && reducemin/ldmin < 1.00001)
  {
    openoutfilemain(k);
    outfilemain << k << " " << min << endl;
    outfilemain.close();
  }
  outfile.close();

  } // end of loop on k

  MPI_Finalize();
}
