#include<iostream>
#include<fstream>
#include "smallprimes.h"
#include "ring.h"
#include "estimate.h"

void sortbyratio(vector<Ring> &r) // uses selection sort
{
  for(int i=0; i+1<r.size(); i++)
  {
    for(int j=i+1; j<r.size(); j++)
      if(r[j].ratio>r[i].ratio)
        { Ring x=r[j]; r[j]=r[i]; r[i]=x; }
  }
}

void print(int k, const vector<Ring> &r, int cutoff) 
{
  char filename[]="xxx.txt";
  filename[0]='0'+k/100;
  filename[1]='0'+(k/10)%10;
  filename[2]='0'+k%10;
  ofstream file(filename);
  file << r.size() << " " << cutoff << " " << r.size()-cutoff << endl << endl;
  for(int i=0; i<r.size(); i++)
  {
    if(i==cutoff) file << endl;
    file << r[i].p << " " << r[i].a << " " << r[i].A << endl;
  }
  file.close();
}

bool unsplit(int k, vector<Ring> &r, int &cutoff)
{
  bool merged=false;
  for(int i=0; i<r.size(); i++)
  {
    // two cases: both below cutoff, or both above
    if(i<cutoff && r[i].split)
    {
      // find our partner
      for(int j=i+1; j<cutoff; j++)
        if(i!=j && r[j].p==r[i].p) // partner found!
        {
          // re-initialize r[i] to the whole ring
          r[i].init(r[j].p,k);
          // slide everyone over
          for(int m=j; m+1<r.size(); m++) r[m]=r[m+1];
          // remove the last item
          r.resize(r.size()-1);
          cutoff--;
          merged=true;
        }
    }
    else if(i>=cutoff && r[i].split)
    {
      // find our partner
      for(int j=i+1; j<r.size(); j++)
        if(i!=j && r[j].p==r[i].p) // partner found!
        {
          // re-initialize r[i] to the whole ring
          r[i].init(r[j].p,k);
          // slide everyone over
          for(int m=j; m+1<r.size(); m++) r[m]=r[m+1];
          // remove the last item
          r.resize(r.size()-1);
          merged=true;
        }
    }
    
  }
  return merged;
}

int main()
{
  int k;

  for(k=351; k<=400; k++)
  {

  double target;
//  cout << "Enter k: ";
//  cin >> k;
  //cout << "Enter g(k) estimate: ";
  //cin >> target;
  target=estimateg(k); // includes fudge factor

  makesmallprimes(k);

  vector<Ring> r;
  r.resize(primelen);

  // first, simple splitting
  for(int i=0; i<primelen; i++)
  {
    r[i].init(prime[i],k);
  }

  // do splitting
  for(int i=0; i<primelen; i++)
  {
    Ring x;
    if(r[i].asplit_simple()<r[i].A)
    {
      // do the split
      x.init(r[i].p,k,r[i].asplit_simple());
      r[i].split=x.split=true;
      r[i].babya=x.a;
      r[i].makeknap();
      r.push_back(r[i]);
      r[i]=x;
    }
  }

  sortbyratio(r);

  double totalsize=0;
  double totalvalue=0;

  double modulus=1;
  double residues=1;
  int cutoff=0;

  for(int i=0; i<r.size(); i++)
  {
    if(totalsize<log(target))
    {
      totalsize+=r[i].size;
      totalvalue+=r[i].value;
      //r[i].print(cout);
    }
    else { cutoff=i; break; } // i and higher are filters instead
  }
  //cout << endl << endl;
  while( unsplit(k,r,cutoff) );
  for(int i=0; i<r.size(); i++)
  {
    //if(i==cutoff) cout << "--"<< endl;
    //r[i].print(cout);
  }
  //print again
/*
  cout << "Total size=" << totalsize
       << " (" << log(target) << ")"
       << " Total value=" << totalvalue
       << " ratio=" << totalvalue/totalsize 
       << endl;
*/

  print(k,r,cutoff);
  cout << k << endl;

  }

  return 0;
}
