#ifndef _RING
#define _RING

#include "int.h"

class Ring
{
  public:

  int16 p; // the prime
  int16 k;
  int16 a; // used for splitting - this ring's a for m=p^a
    // if a!=A, then this prime was split
  int16 babya; // the a value of the other half of the split
  int32 m; // p^a

  // representing k in base p:
  int16 dlen; // min such that p^dlen >= k
  int16 A; // (alias didn't work) alias for dlen
  vector<int16> d; // k in base p

  // bit vector for sieving/wheel
  vector<bool> ones; // sieving array
  int32 numres; // number of ones in the ones vector

  // r, R vectors used in computing optimal splits
  vector<int16> r;
  vector<int32> R;  // r and R arrays from the writeup on splitting

  // knapsack data
  double size, value, ratio; // for knapsack

  bool split; // true only for rings from a split

  // initiallizer functions for the data

  void maked() //assumes p,k initialized, computes A,d[]
  {
    int kin=k;
    dlen=0; d.clear(); d.resize(int(log(k)/log(p))+4,0); // plenty of room
    while(kin>0) { d[dlen++]=kin%p; kin=kin/p; }
    A=dlen;
  }

  void makem() //computes m from p, a
  {
    m=1;
    for(int i=0; i<a; i++) m*=p;
  }

private:
  bool isbigger(int32 x, int16 k) // used by makeones()
  {
    k=k%m; x=x%m;
    while(x>0 && k>0)
    {
      if(x%p<k%p) return false;
      x=x/p; k=k/p;
    }
    return (x>=k);
  }

public:
  void makeones() //computes ones[], numres from k,p,m (makem() called)
  {
    ones.clear();
    ones.resize(m+1,false);
    numres=0;
    for(int32 i=0; i<m; i++)
    {
      ones[i]=isbigger(i,k);
      numres+=(ones[i]?1:0);
    }
  }

  void makerR() // assumes maked() has been called already
  {
    r.resize(A,0);
    R.resize(A+1,1);
    R[0]=1;
    for(int i=0; i<A; i++)
    {
      r[i]=p-d[i];
      R[i+1]=R[i]*r[i];
    }
  }

  void makeknap() // knapsack data - makeones(), makem() already called
  {
    value=log((1.0*m)/numres);
    size=log(m);
    ratio=value/size;

    if(!split) return;
    if(a<A) return;
    makerR();
    size=(A-babya)*log(p);
    value=size-log(R[A]/R[babya]);
    ratio=value/size;
  }

  //other functions

  Ring& operator=(const Ring& x)
  {
    init(x.p,x.k,x.a);
    split=x.split;
    babya=x.babya;
    makeknap();
    return *this;
  }

  void init(int16 pin, int16 kin)  // for setting a=A
  {
    p=pin; k=kin;
    maked(); a=A;
    makem(); makeones(); makeknap(); makerR();
    split=false;
  }

  void init(int16 pin, int16 kin, int16 ain)
  {
    p=pin; k=kin; a=ain;
    maked(); makem(); makeones(); makeknap(); makerR();
    split=true;
  }

  int16 asplit_simple() // uses the r,R arrays
  {
    double minval=log(R[1]);
    int16 mina=1;
    for(int a=2; a<=A; a++)
    {
      if(log(R[a])/(1.0*a) <= minval)
      { mina=a; minval=log(R[a])/a; }
    }
    return mina;
  }

  //int16 asplit_weighted(double sz, double vl)
  int16 asplit(double rat)  // so rat=val/size<=1
  {
    // first compute log(s)
    double logs= log(p)*(1-rat);

    double minval=log(R[1])+(A-1)*logs;
    int16 mina=1;
    for(int a=2; a<=A; a++)
    {
      if(log(R[a])+(A-a)*logs <= minval)
      { mina=a; minval=log(R[a])+(A-a)*logs; }
    }
    return mina;
    
  }


public:

  void print(ostream& os)
  {
    os << p  
       << " m=" << m
       << " a=" << a
       << " A=" << A
       << (split?" SPLIT ":" ")
       << " \tk base p = ";
    for(int i=dlen-1; i>=0; i--) os << d[i] << " "; os << endl;
    os << "simple split=" << asplit_simple()
       << " numres=" << numres
       << " R[A]=" << R[A]
       << endl;
    os << " value=" << value
       << " size=" << size
       << " ratio=" << ratio
       << endl;

    if(m<100)
    {
      for(int j=0; j<m; j++) os << (ones[j]?1:0) ;
      os << endl;
    }
  }

};

#endif
