#ifndef _SPOKE
#define _SPOKE

#include<iostream>
#include "int.h"
#include "bigint.h"
using namespace std;

class Spoke
{
  public:
  int128 m; // input modulus (product of previous primes)
  int32 p, mmodp; // the prime (or product thereof)
  int32 pos; // current squares position for iteration
  int128 val; // current squares value for iteration
  int32 numones; // number of one bits
  int32 numleft; // number left to iterate (count down)
  vector<int128> jump;
  vector<int32> jmodp;
  vector<bool> isone;

  inline int128 first(int128 x) 
  { 
    pos=x%p; val=x; numleft=numones-1; 
    if(!isone[pos])
      { val+=jump[pos]; pos=(pos+jmodp[pos])%p; }
    return val;
  }

  inline int128 next() 
  { 
    numleft--;
    val += jump[pos];
    pos = (pos+jmodp[pos])%p;
    return val;
  }

  inline bool hasNext() { return (numleft>0);}

  void init(int32 pin, int128 min, const vector<bool> & ones)
  {
    p=pin; m=min;
    mmodp=m%p;
    numones=0;
    isone.resize(p+1);
    for(int i=0; i<p; i++) 
      { isone[i]=ones[i]; if(ones[i]) numones++; }
    dojumps();
  }

  void print(ostream & os)
  {
    os << "p=" << p << " m=" << m
       << " m mod p=" << mmodp << endl;
    for(int i=0; i<p; i++)
      os << (isone[i]?'1':'0');
    os << endl;
    for(int i=0; i<p; i++)
      os << jump[i] << " ";
    os << endl;
    for(int i=0; i<p; i++)
      os << jmodp[i] << " ";
    os << endl << endl;
  }
  
  private:
  void dojumps()
  {
    jump.resize(p+1);
    jmodp.resize(p+1);
    for(int j=0; j<p; j++)
      for(int k=1; k<p; k++)
        if( isone[ (j+k*mmodp)%p ] )
        {
           jump[j]=k*m; jmodp[j]=(k*mmodp)%p;
           break;
        }
  }
};

#endif
