#ifndef _WHEEL
#define _WHEEL

#include<iostream>
#include<fstream>
#include<iomanip>
#include<string>
#include "spoke.h"

#include "ring.h"

using namespace std;

class Wheel
{
  public:
  vector<Spoke> s;
  int nums;
  int128 m;
  int128 numones;
  int128 max;

  public:

  void print(ostream &out)
  {
    out << "Wheel: size="<<nums<<" modulus="<<m<<" numones="<<numones<<endl;
    for(int i=0; i<nums; i++)
      s[i].print(out);
    out << endl;
  }

  void build2(const vector<Ring> &r, int start, int cutoff, int128 maxin,
    int128 min)
  {
    m=min;
    nums=0;
    numones=1;
    max=maxin;
    s.resize(cutoff);
    vector<bool> ones;
    for(int i=start; i<cutoff; i++)
    {
      s[nums++].init(r[i].m,m,r[i].ones);
      m*=r[i].m;
      numones*= s[nums-1].numones;
    }
  }
  int build1(const vector<Ring> &r, int cutoff, int128 maxin, int128 processors)
  // returns start value for second wheel build
  // makes sure numones is >= processors
  {
    m=1;
    nums=0;
    numones=1;
    max=maxin;
    s.resize(cutoff);
    vector<bool> ones;
    for(int i=0; i<cutoff; i++)
    {
      s[nums++].init(r[i].m,m,r[i].ones);
      m*=r[i].m;
      numones*= s[nums-1].numones;
      if(numones>=processors) { return i+1; }
    }
    return cutoff;
  }

  void build(const vector<Ring> &r, int cutoff, int128 maxin)
  {
    m=1;
    nums=0;
    numones=1;
    max=maxin;
    s.resize(cutoff);
    vector<bool> ones;
    for(int i=0; i<cutoff; i++)
    {
      s[nums++].init(r[i].m,m,r[i].ones);
      m*=r[i].m;
      numones*= s[nums-1].numones;
    }
  }

  int128 first(int128 x)
  {
    for(int i=0; i<nums; i++)
      { x=s[i].first(x); }
    return x;
  }

/*
  int128 next()
  {
    int i=nums-1;
    while(i>=0 && !s[i].hasNext()) i--;
    int128 x=s[i].next(); i++;
    for(; i<nums; i++) { x=s[i].first(x); }
    return x;
  }
/*

*/
  int128 next() // version with max limit imposed
  {
    // first, handle the most common situation for speed
    int128 x;
    int level=nums-1;
    if(s[level].hasNext())
    {
      x=s[level].next();
      if(x<=max) return x;
    }

    do {
      // if we get here, we're done with this level
      level--;
      // back up to find a next item
      while(level>=0 && !s[level].hasNext()) level--;
      if(level<0) return -1; // we finished the wheel!
      x=s[level].next(); level++;
      for(; level<nums && x<=max; level++) { x=s[level].first(x); }

    } while(x>max); // end of do-while loop
    return x;
  }
/*
*/

  bool hasNext()
  {
    for(int i=nums-1; i>=0; i--)
      if(s[i].hasNext()) return true;
    return false;
  }

  void odometer(ostream &os)
  {
    for(int i=0; i<nums; i++)
      os << setw(3) << s[i].numleft << "/" << s[i].numones;
    os << endl;
  }

};

#endif
